@@ -1,7 +1,7 @@
 /*
  * This program attempts to recover deleted data structures in a registry hive.
  *
- * Copyright (C) 2008 Timothy D. Morgan
+ * Copyright (C) 2008-2009 Timothy D. Morgan
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,18 +14,18 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
+ * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * $Id: reglookup-recover.c 121 2008-08-09 17:22:26Z tim $
+ * $Id: reglookup-recover.c 152 2009-06-02 20:00:38Z tim $
  */
 
 #include <stdio.h>
 #include <stdlib.h>
-#include <sysexits.h>
 
-#include "../include/regfi.h"
-#include "../include/range_list.h"
-#include "../include/lru_cache.h"
+#include "talloc.h"
+#include "regfi.h"
+#include "range_list.h"
+#include "lru_cache.h"
 
 
 /* Globals, influenced by command line parameters */
@@ -66,7 +66,7 @@ char* getQuotedData(int fd, uint32 offset, uint32 length)
 }
 
 
-void printKey(REGF_FILE* f, REGF_NK_REC* nk, const char* prefix)
+void printKey(REGFI_FILE* f, REGFI_NK_REC* nk, const char* prefix)
 {
   char mtime[20];
   time_t tmp_time[1];
@@ -83,10 +83,11 @@ void printKey(REGF_FILE* f, REGF_NK_REC* nk, const char* prefix)
   {
     quoted_name = malloc(1*sizeof(char));
     if(quoted_name == NULL)
-      bailOut(EX_OSERR, "ERROR: Could not allocate sufficient memory.\n");
+      bailOut(REGLOOKUP_EXIT_OSERR, 
+	      "ERROR: Could not allocate sufficient memory.\n");
     quoted_name[0] = '\0';
 
-    fprintf(stderr, "WARNING: NULL key name in NK record at offset %.8X.\n",
+    fprintf(stderr, "WARN: NULL key name in NK record at offset %.8X.\n",
 	    nk->offset);
   }
 
@@ -98,10 +99,11 @@ void printKey(REGF_FILE* f, REGF_NK_REC* nk, const char* prefix)
   
   if(print_parsedraw)
     free(quoted_raw);
+  free(quoted_name);
 }
 
 
-void printValue(REGF_FILE* f, const REGF_VK_REC* vk, const char* prefix)
+void printValue(REGFI_FILE* f, const REGFI_VK_REC* vk, const char* prefix)
 {
   char* quoted_value = NULL;
   char* quoted_name = NULL;
@@ -121,11 +123,11 @@ void printValue(REGF_FILE* f, const REGF_VK_REC* vk, const char* prefix)
    *      name of the current value we're having trouble with, since
    *      stderr/stdout don't always sync nicely.
    */
-  if(size > VK_MAX_DATA_LENGTH)
+  if(size > REGFI_VK_MAX_DATA_LENGTH)
   {
-    fprintf(stderr, "WARNING: value data size %d larger than "
-	    "%d, truncating...\n", size, VK_MAX_DATA_LENGTH);
-    size = VK_MAX_DATA_LENGTH;
+    fprintf(stderr, "WARN: value data size %d larger than "
+	    "%d, truncating...\n", size, REGFI_VK_MAX_DATA_LENGTH);
+    size = REGFI_VK_MAX_DATA_LENGTH;
   }
   
   quoted_name = quote_string(vk->valuename, key_special_chars);
@@ -138,7 +140,7 @@ void printValue(REGF_FILE* f, const REGF_VK_REC* vk, const char* prefix)
      */
     quoted_name = malloc(1*sizeof(char));
     if(quoted_name == NULL)
-      bailOut(EX_OSERR, "ERROR: Could not allocate sufficient memory.\n");
+      bailOut(REGLOOKUP_EXIT_OSERR, "ERROR: Could not allocate sufficient memory.\n");
     quoted_name[0] = '\0';
   }
 
@@ -147,19 +149,19 @@ void printValue(REGF_FILE* f, const REGF_VK_REC* vk, const char* prefix)
   {
     quoted_value = malloc(1*sizeof(char));
     if(quoted_value == NULL)
-      bailOut(EX_OSERR, "ERROR: Could not allocate sufficient memory.\n");
+      bailOut(REGLOOKUP_EXIT_OSERR, "ERROR: Could not allocate sufficient memory.\n");
     quoted_value[0] = '\0';
 
     if(conv_error == NULL)
-      fprintf(stderr, "WARNING: Could not quote value for '%s/%s'.  "
+      fprintf(stderr, "WARN: Could not quote value for '%s/%s'.  "
 	      "Memory allocation failure likely.\n", prefix, quoted_name);
     else if(print_verbose)
-      fprintf(stderr, "WARNING: Could not quote value for '%s/%s'.  "
+      fprintf(stderr, "WARN: Could not quote value for '%s/%s'.  "
 	      "Returned error: %s\n", prefix, quoted_name, conv_error);
   }
   /* XXX: should these always be printed? */
   else if(conv_error != NULL && print_verbose)
-    fprintf(stderr, "VERBOSE: While quoting value for '%s/%s', "
+    fprintf(stderr, "INFO: While quoting value for '%s/%s', "
 	    "warning returned: %s\n", prefix, quoted_name, conv_error);
 
 
@@ -187,7 +189,7 @@ void printValue(REGF_FILE* f, const REGF_VK_REC* vk, const char* prefix)
 }
 
 
-void printSK(REGF_FILE* f, REGF_SK_REC* sk)
+void printSK(REGFI_FILE* f, REGFI_SK_REC* sk)
 {
   char* quoted_raw = NULL;
   char* empty_str = "";
@@ -225,7 +227,7 @@ void printSK(REGF_FILE* f, REGF_SK_REC* sk)
 }
 
 
-int printCell(REGF_FILE* f, uint32 offset)
+int printCell(REGFI_FILE* f, uint32 offset)
 {
   char* quoted_buf;
   uint32 cell_length;
@@ -250,65 +252,75 @@ int printCell(REGF_FILE* f, uint32 offset)
  * Paths returned must be free()d.
  */
 /* XXX: This is not terribly efficient, as it may reparse many keys 
- *      repeatedly.  Should try to add caching.  Also, piecing the path 
- *      together is slow and redundant.
+ *      repeatedly.  Should try to add caching.
  */
-char* getParentPath(REGF_FILE* f, REGF_NK_REC* nk)
+char* getParentPath(REGFI_FILE* f, REGFI_NK_REC* nk)
 {
-  void_stack* path_stack = void_stack_new(REGF_MAX_DEPTH);
-  REGF_HBIN* hbin;
-  REGF_NK_REC* cur_ancestor;
+  void_stack* path_stack = void_stack_new(REGFI_MAX_DEPTH);
+  const REGFI_HBIN* hbin;
+  REGFI_NK_REC* cur_ancestor;
   char* ret_val;
-  char* path_element;
-  char* tmp_str;
-  uint32 virt_offset, i, stack_size, ret_val_size, ret_val_left, element_size;
+  uint32 virt_offset, i, stack_size, ret_val_size, ret_val_used;
   uint32 max_length;
-
+  REGFI_BUFFER* path_element;
+  
   /* The path_stack size limit should guarantee that we don't recurse forever. */
   virt_offset = nk->parent_off;
-  while(virt_offset != REGF_OFFSET_NONE)
+  ret_val_size = 1; /* NUL */
+  while(virt_offset != REGFI_OFFSET_NONE)
   {  
     hbin = regfi_lookup_hbin(f, virt_offset);
     if(hbin == NULL)
-      virt_offset = REGF_OFFSET_NONE;
+      virt_offset = REGFI_OFFSET_NONE;
     else
     {
       max_length = hbin->block_size + hbin->file_off 
-	- (virt_offset+REGF_BLOCKSIZE);
-      cur_ancestor = regfi_parse_nk(f, virt_offset+REGF_BLOCKSIZE, 
+	- (virt_offset+REGFI_REGF_SIZE);
+      cur_ancestor = regfi_parse_nk(f, virt_offset+REGFI_REGF_SIZE, 
 				    max_length, true);
+      printMsgs(f);
+
       if(cur_ancestor == NULL)
-	virt_offset = REGF_OFFSET_NONE;
+	virt_offset = REGFI_OFFSET_NONE;
       else
       {
-	if(cur_ancestor->key_type == NK_TYPE_ROOTKEY)
-	  virt_offset = REGF_OFFSET_NONE;
+	if(cur_ancestor->key_type & REGFI_NK_FLAG_ROOT)
+	  virt_offset = REGFI_OFFSET_NONE;
 	else
 	  virt_offset = cur_ancestor->parent_off;
 	
-	path_element = quote_string(cur_ancestor->keyname, key_special_chars);
-	if(path_element == NULL || !void_stack_push(path_stack, path_element))
+	path_element = talloc(path_stack, REGFI_BUFFER);
+	if(path_element != NULL)
+	  path_element->buf = (uint8*)quote_string(cur_ancestor->keyname, 
+						   key_special_chars);
+	  
+	if(path_element == NULL || path_element->buf == NULL 
+	   || !void_stack_push(path_stack, path_element))
 	{
-	  free(cur_ancestor->keyname);
-	  free(cur_ancestor);
-	  void_stack_free_deep(path_stack);
+	  /* XXX: Need to add a warning here */
+	  regfi_free_key(cur_ancestor);
+	  void_stack_free(path_stack);
 	  return NULL;
 	}
 
-	regfi_key_free(cur_ancestor);
+	/* Path element and preceeding delimiter
+	 * Note that this integer can't overflow since key name lengths are
+	 * 16 bits and the max depth is 512.
+	 */
+	path_element->len = strlen((char*)path_element->buf);
+	ret_val_size += path_element->len + 1;
+
+	regfi_free_key(cur_ancestor);
       }
     }
   }
   
   stack_size = void_stack_size(path_stack);
-  ret_val_size = 16*stack_size;
-  if(ret_val_size == 0)
-    ret_val_size = 1;
-  ret_val_left = ret_val_size;
+  ret_val_used = 0;
   ret_val = malloc(ret_val_size);
   if(ret_val == NULL)
   {
-    void_stack_free_deep(path_stack);
+    void_stack_free(path_stack);
     return NULL;
   }
   ret_val[0] = '\0';
@@ -316,23 +328,11 @@ char* getParentPath(REGF_FILE* f, REGF_NK_REC* nk)
   for(i=0; i<stack_size; i++)
   {
     path_element = void_stack_pop(path_stack);
-    element_size = strlen(path_element);
-    if(ret_val_left < element_size+2)
-    {
-      ret_val_size += element_size+16;
-      ret_val_left += element_size+16;
-      tmp_str = (char*)realloc(ret_val, ret_val_size);
-      if(tmp_str == NULL)
-      {
-	free(ret_val);
-	void_stack_free_deep(path_stack);
-	return NULL;
-      }
-      ret_val = tmp_str;
-    }
-
-    ret_val_left -= snprintf(ret_val+ret_val_size-ret_val_left,ret_val_left, "/%s", path_element);
-    free(path_element);
+    snprintf(ret_val+ret_val_used, ret_val_size-ret_val_used, 
+	     "/%s", path_element->buf);
+    ret_val_used += path_element->len + 1;
+    free(path_element->buf);
+    talloc_free(path_element);
   }
   void_stack_free(path_stack);
 
@@ -412,183 +412,24 @@ bool removeRange(range_list* rl, uint32 offset, uint32 length)
 }
 
 
-/* NOTE: unalloc_keys should be an empty range_list. */
-int extractKeys(REGF_FILE* f, 
-		range_list* unalloc_cells, 
-		range_list* unalloc_keys)
+int extractVKs(REGFI_FILE* f,
+	       range_list* unalloc_cells,
+	       range_list* unalloc_values)
 {
   const range_list_element* cur_elem;
-  REGF_NK_REC* key;
+  REGFI_VK_REC* vk;
   uint32 i, j;
 
   for(i=0; i < range_list_size(unalloc_cells); i++)
   {
-    cur_elem = range_list_get(unalloc_cells, i);
-    for(j=0; cur_elem->length > REGFI_NK_MIN_LENGTH 
-	  && j <= cur_elem->length-REGFI_NK_MIN_LENGTH; j+=8)
-    {
-      key = regfi_parse_nk(f, cur_elem->offset+j,
-			   cur_elem->length-j, false);
-      if(key != NULL)
-      {
-	if(!range_list_add(unalloc_keys, key->offset, 
-			   key->cell_size, key))
-	{
-	  fprintf(stderr, "ERROR: Couldn't add key to unalloc_keys.\n");
-	  return 20;
-	}
-	j+=key->cell_size-8;
-      }
-    }
-  }
-
-  for(i=0; i<range_list_size(unalloc_keys); i++)
-  {
-    cur_elem = range_list_get(unalloc_keys, i);
-    if(!removeRange(unalloc_cells, cur_elem->offset, cur_elem->length))
-      return 30;
-  }
-
-  return 0;
-}
-
-
-int extractValueLists(REGF_FILE* f,
-		      range_list* unalloc_cells,
-		      range_list* unalloc_keys)
-{
-  REGF_NK_REC* nk;
-  REGF_HBIN* hbin;
-  const range_list_element* cur_elem;
-  uint32 i, j, num_keys, off, values_length, max_length;
-
-  num_keys=range_list_size(unalloc_keys);
-  for(i=0; i<num_keys; i++)
-  {
-    cur_elem = range_list_get(unalloc_keys, i);
-    if(cur_elem == NULL)
-      return 10;
-    nk = cur_elem->data;
-
-    if(nk->num_values && (nk->values_off!=REGF_OFFSET_NONE))
-    {
-      hbin = regfi_lookup_hbin(f, nk->values_off);
-      
-      if(hbin != NULL)
-      {
-	off = nk->values_off + REGF_BLOCKSIZE;
-	max_length = hbin->block_size + hbin->file_off - off;
-	/* XXX: This is a hack.  We parse all value-lists, VK records,
-	 *      and data records without regard for current allocation status.  
-	 *      On the off chance that such a record correctly parsed but is 
-	 *      actually a reallocated structure used by something else, we 
-	 *      simply prune it after the fact.  Would be faster to check this
-	 *      up front somehow.
-	 */
-	nk->values = regfi_load_valuelist(f, off, nk->num_values, max_length,
-					  false);
-	values_length = (nk->num_values+1)*sizeof(uint32);
-	if(values_length != (values_length & 0xFFFFFFF8))
-	  values_length = (values_length & 0xFFFFFFF8) + 8;
-
-	if(nk->values != NULL)
-	{
-	  if(!range_list_has_range(unalloc_cells, off, values_length))
-	  { /* We've parsed a values-list which isn't in the unallocated list,
-	     * so prune it. 
-	     */
-	    for(j=0; j<nk->num_values; j++)
-	    {
-	      if(nk->values[j] != NULL)
-	      {
-		if(nk->values[j]->data != NULL)
-		  free(nk->values[j]->data);
-		free(nk->values[j]);
-	      }
-	    }
-	    free(nk->values);
-	    nk->values = NULL;
-	  }
-	  else
-	  { /* Values-list was recovered.  Remove from unalloc_cells and 
-	     * inspect values. 
-	     */
-	    if(!removeRange(unalloc_cells, off, values_length))
-	      return 20;
-
-	    for(j=0; j < nk->num_values; j++)
-	    {
-	      if(nk->values[j] != NULL)
-	      {
-		if(!range_list_has_range(unalloc_cells, nk->values[j]->offset, 
-					 nk->values[j]->cell_size))
-		{ /* We've parsed a value which isn't in the unallocated list,
-		   * so prune it.
-		   */
-		  if(nk->values[j]->data != NULL)
-		    free(nk->values[j]->data);
-		  free(nk->values[j]);
-		  nk->values[j] = NULL;
-		}
-		else
-		{
-		  /* A VK record was recovered.  Remove from unalloc_cells
-		   * and inspect data.
-		   */
-		  if(!removeRange(unalloc_cells, nk->values[j]->offset,
-				  nk->values[j]->cell_size))
-		    return 21;
-
-		  /* Don't bother pruning or removing from unalloc_cells if 
-		   * there is no data, or it is stored in the offset.
-		   */
-		  if(nk->values[j]->data != NULL && !nk->values[j]->data_in_offset)
-		  {
-		    off = nk->values[j]->data_off+REGF_BLOCKSIZE;
-		    if(!range_list_has_range(unalloc_cells, off, 
-					     nk->values[j]->data_size))
-		    { /* We've parsed a data cell which isn't in the unallocated 
-		       * list, so prune it.
-		       */
-		      free(nk->values[j]->data);
-		      nk->values[j]->data = NULL;
-		    }
-		    else
-		    { /*A data record was recovered. Remove from unalloc_cells.*/
-		      if(!removeRange(unalloc_cells, off, 
-				      nk->values[j]->data_size))
-			return 22;
-		    }
-		  }
-		}
-	      }
-	    }
-	  }
-	}
-      }
-    }
-  }
-
-  return 0;
-}
-
-
-/* NOTE: unalloc_values should be an empty range_list. */
-int extractValues(REGF_FILE* f,
-		  range_list* unalloc_cells,
-		  range_list* unalloc_values)
-{
-  const range_list_element* cur_elem;
-  REGF_VK_REC* vk;
-  uint32 i, j, off;
-
-  for(i=0; i < range_list_size(unalloc_cells); i++)
-  {
+    printMsgs(f);
     cur_elem = range_list_get(unalloc_cells, i);
     for(j=0; j <= cur_elem->length; j+=8)
     {
       vk = regfi_parse_vk(f, cur_elem->offset+j, 
 			   cur_elem->length-j, false);
+      printMsgs(f);
+
       if(vk != NULL)
       {
 	if(!range_list_add(unalloc_values, vk->offset,
@@ -601,7 +442,7 @@ int extractValues(REGF_FILE* f,
       }
     }
   }
-  
+
   /* Remove value ranges from the unalloc_cells before we continue. */
   for(i=0; i<range_list_size(unalloc_values); i++)
   {
@@ -610,28 +451,76 @@ int extractValues(REGF_FILE* f,
       return 30;
   }
 
-  /* Now see if the data associated with each value is intact */
+  return 0;
+}
+
+
+int extractDataCells(REGFI_FILE* f,
+		     range_list* unalloc_cells,
+		     range_list* unalloc_values)
+{
+  const range_list_element* cur_elem;
+  REGFI_VK_REC* vk;
+  const REGFI_HBIN* hbin;
+  REGFI_BUFFER data;
+  uint32 i, off, data_offset, data_maxsize;
+
   for(i=0; i<range_list_size(unalloc_values); i++)
   {
     cur_elem = range_list_get(unalloc_values, i);
-    vk = (REGF_VK_REC*)cur_elem->data;
+    vk = (REGFI_VK_REC*)cur_elem->data;
     if(vk == NULL)
       return 40;
 
-    if(vk->data != NULL && !vk->data_in_offset)
+    if(vk->data_size == 0)
+      vk->data = NULL;
+    else
     {
-      off = vk->data_off+REGF_BLOCKSIZE;
-      if(!range_list_has_range(unalloc_cells, off, vk->data_size))
-      { /* We've parsed a data cell which isn't in the unallocated 
-	 * list, so prune it.
-	 */
-	free(vk->data);
-	vk->data = NULL;
+      off = vk->data_off+REGFI_REGF_SIZE;
+
+      if(vk->data_in_offset)
+      {
+	data = regfi_load_data(f, vk->type, vk->data_off,
+			       vk->data_size, 4,
+			       vk->data_in_offset, false);
+	vk->data = data.buf;
+	vk->data_size = data.len;
       }
-      else
-      { /*A data record was recovered. Remove from unalloc_cells.*/
-	if(!removeRange(unalloc_cells, off, vk->data_size))
-	  return 50;
+      else if(range_list_has_range(unalloc_cells, off, vk->data_size))
+      {
+	hbin = regfi_lookup_hbin(f, vk->data_off);
+	if(hbin)
+	{
+	  data_offset = vk->data_off+REGFI_REGF_SIZE;
+	  data_maxsize = hbin->block_size + hbin->file_off - data_offset;
+	  data = regfi_load_data(f, vk->type, data_offset, 
+				 vk->data_size, data_maxsize, 
+				 vk->data_in_offset, false);
+	  vk->data = data.buf;
+	  vk->data_size = data.len;
+
+	  if(vk->data != NULL)
+	  {
+	    /* XXX: The following may not make sense now in light of big data
+	     *      records.
+	     */
+	    /* XXX: This strict checking prevents partial recovery of data 
+	     *      cells.  Also, see code for regfi_load_data and note that
+	     *      lengths indicated in VK records are sometimes just plain 
+	     *      wrong.  Need a feedback mechanism to be more fuzzy with 
+	     *      data cell lengths and the ranges removed. 
+	     *
+	     *      The introduction of REGFI_BUFFER in regfi_load_data has 
+	     *      fixed some of this.  Should review again with respect to 
+	     *      the other issues mentioned above though.
+	     */
+	    /* A data record was recovered. Remove from unalloc_cells. */
+	    if(!removeRange(unalloc_cells, off, vk->data_size))
+	      return 50;
+	  }
+	}
+	else
+	  vk->data = NULL;
       }
     }
   }
@@ -640,22 +529,171 @@ int extractValues(REGF_FILE* f,
 }
 
 
+/* NOTE: unalloc_keys should be an empty range_list. */
+int extractKeys(REGFI_FILE* f, 
+		range_list* unalloc_cells, 
+		range_list* unalloc_keys)
+{
+  const range_list_element* cur_elem;
+  REGFI_NK_REC* key;
+  uint32 i, j;
+  int error_code = 0;
+
+  for(i=0; i < range_list_size(unalloc_cells); i++)
+  {
+    printMsgs(f);
+    cur_elem = range_list_get(unalloc_cells, i);
+    for(j=0; cur_elem->length > REGFI_NK_MIN_LENGTH 
+	  && j <= cur_elem->length-REGFI_NK_MIN_LENGTH; j+=8)
+    {
+      key = regfi_parse_nk(f, cur_elem->offset+j,
+			   cur_elem->length-j, false);
+      printMsgs(f);
+
+      if(key != NULL)
+      {
+	if(!range_list_add(unalloc_keys, key->offset, 
+			   key->cell_size, key))
+	{
+	  fprintf(stderr, "ERROR: Couldn't add key to unalloc_keys.\n");
+	  error_code = 20;
+	  goto fail;
+	}
+	talloc_steal(unalloc_keys, key);
+	j+=key->cell_size-8;
+      }
+    }
+  }
+
+  for(i=0; i<range_list_size(unalloc_keys); i++)
+  {
+    cur_elem = range_list_get(unalloc_keys, i);
+    if(!removeRange(unalloc_cells, cur_elem->offset, cur_elem->length))
+    {
+      error_code = 30;
+      goto fail;
+    }
+  }
+
+  return 0;
+
+ fail:
+  regfi_free_key(key);
+  return error_code;
+}
+
+int extractValueLists(REGFI_FILE* f,
+		      range_list* unalloc_cells,
+		      range_list* unalloc_keys,
+		      range_list* unalloc_linked_values)
+{
+  REGFI_NK_REC* nk;
+  REGFI_VK_REC* vk;
+  const REGFI_HBIN* hbin;
+  const range_list_element* cur_elem;
+  uint32 i, j, num_keys, off, values_length, max_length;
+
+  num_keys=range_list_size(unalloc_keys);
+  for(i=0; i<num_keys; i++)
+  {
+    cur_elem = range_list_get(unalloc_keys, i);
+    if(cur_elem == NULL)
+      return 10;
+    nk = cur_elem->data;
+
+    if(nk->num_values && (nk->values_off!=REGFI_OFFSET_NONE))
+    {
+      hbin = regfi_lookup_hbin(f, nk->values_off);
+      
+      if(hbin != NULL)
+      {
+	off = nk->values_off + REGFI_REGF_SIZE;
+	max_length = hbin->block_size + hbin->file_off - off;
+	nk->values = regfi_load_valuelist(f, off, nk->num_values, 
+					  max_length, false);
+	if(nk->values != NULL && nk->values->elements != NULL)
+	{
+	  /* Number of elements in the value list may be shorter than advertised 
+	   * by NK record due to cell truncation.  We'll consider this valid and 
+	   * only throw out the whole value list if it bleeds into an already 
+	   * parsed structure.
+	   */
+	  values_length = (nk->values->num_values+1)*sizeof(uint32);
+	  if(values_length != (values_length & 0xFFFFFFF8))
+	    values_length = (values_length & 0xFFFFFFF8) + 8;
+
+	  if(!range_list_has_range(unalloc_cells, off, values_length))
+	  { /* We've parsed a values-list which isn't in the unallocated list,
+	     * so prune it.
+	     */
+	    talloc_free(nk->values);
+	    nk->values = NULL;
+	  }
+	  else
+	  { /* Values-list was recovered.  Remove from unalloc_cells and 
+	     * inspect values. 
+	     */
+	    if(!removeRange(unalloc_cells, off, values_length))
+	      return 20;
+
+	    for(j=0; j < nk->values->num_values; j++)
+	    {
+	      /* Don't bother to restrict cell length here, since we'll
+	       * check our unalloc_cells range_list later. 
+	       */
+	      vk = regfi_parse_vk(f, nk->values->elements[j]+REGFI_REGF_SIZE,
+				  0x7FFFFFFF, false);
+	      printMsgs(f);
+	      
+	      if(vk != NULL)
+	      {
+		if(range_list_has_range(unalloc_cells, 
+					vk->offset, vk->cell_size))
+		{
+		  if(!range_list_add(unalloc_linked_values, vk->offset,
+				     vk->cell_size, vk))
+		  {
+		    talloc_free(vk);
+		    return 30;
+		  }
+
+		  if(!removeRange(unalloc_cells, vk->offset, vk->cell_size))
+		    return 40;
+		}
+		else
+		  talloc_free(vk);
+	      }
+	    }
+	  }
+	}
+      }
+    }
+  }
+
+  return 0;
+}
+
+
+
 /* NOTE: unalloc_sks should be an empty range_list. */
-int extractSKs(REGF_FILE* f, 
+int extractSKs(REGFI_FILE* f, 
 	       range_list* unalloc_cells,
 	       range_list* unalloc_sks)
 {
   const range_list_element* cur_elem;
-  REGF_SK_REC* sk;
+  REGFI_SK_REC* sk;
   uint32 i, j;
 
   for(i=0; i < range_list_size(unalloc_cells); i++)
   {
+    printMsgs(f);
     cur_elem = range_list_get(unalloc_cells, i);
     for(j=0; j <= cur_elem->length; j+=8)
     {
       sk = regfi_parse_sk(f, cur_elem->offset+j, 
 			  cur_elem->length-j, false);
+      printMsgs(f);
+
       if(sk != NULL)
       {
 	if(!range_list_add(unalloc_sks, sk->offset,
@@ -664,6 +702,7 @@ int extractSKs(REGF_FILE* f,
 	  fprintf(stderr, "ERROR: Couldn't add sk to unalloc_sks.\n");
 	  return 20;
 	}
+	talloc_steal(unalloc_sks, sk);
 	j+=sk->cell_size-8;
       }
     }
@@ -682,25 +721,25 @@ int extractSKs(REGF_FILE* f,
 
 int main(int argc, char** argv)
 { 
-  REGF_FILE* f;
+  REGFI_FILE* f;
   const range_list_element* cur_elem;
   range_list* unalloc_cells;
   range_list* unalloc_keys;
+  range_list* unalloc_linked_values;
   range_list* unalloc_values;
   range_list* unalloc_sks;
   char** parent_paths;
   char* tmp_name;
   char* tmp_path;
-  REGF_NK_REC* tmp_key;
-  REGF_VK_REC* tmp_value;
+  REGFI_NK_REC* tmp_key;
+  REGFI_VK_REC* tmp_value;
   uint32 argi, arge, i, j, ret, num_unalloc_keys;
-  /* uint32 test_offset;*/
   
   /* Process command line arguments */
   if(argc < 2)
   {
     usage();
-    bailOut(EX_USAGE, "ERROR: Requires at least one argument.\n");
+    bailOut(REGLOOKUP_EXIT_USAGE, "ERROR: Requires at least one argument.\n");
   }
   
   arge = argc-1;
@@ -724,20 +763,24 @@ int main(int argc, char** argv)
     {
       usage();
       fprintf(stderr, "ERROR: Unrecognized option: %s\n", argv[argi]);
-      bailOut(EX_USAGE, "");
+      bailOut(REGLOOKUP_EXIT_USAGE, "");
     }
   }
   /*test_offset = strtol(argv[argi++], NULL, 16);*/
 
   if((registry_file = strdup(argv[argi])) == NULL)
-    bailOut(EX_OSERR, "ERROR: Memory allocation problem.\n");
+    bailOut(REGLOOKUP_EXIT_OSERR, "ERROR: Memory allocation problem.\n");
 
   f = regfi_open(registry_file);
   if(f == NULL)
   {
     fprintf(stderr, "ERROR: Couldn't open registry file: %s\n", registry_file);
-    bailOut(EX_NOINPUT, "");
+    bailOut(REGLOOKUP_EXIT_NOINPUT, "");
   }
+  if(print_verbose)
+    regfi_set_message_mask(f, REGFI_MSG_ERROR|REGFI_MSG_WARN|REGFI_MSG_INFO);
+  else
+    regfi_set_message_mask(f, REGFI_MSG_ERROR);
 
   if(print_header)
     printf("OFFSET,REC_LENGTH,REC_TYPE,PATH,NAME,"
@@ -755,6 +798,10 @@ int main(int argc, char** argv)
   if(unalloc_keys == NULL)
     return 10;
 
+  unalloc_linked_values = range_list_new();
+  if(unalloc_linked_values == NULL)
+    return 10;
+
   unalloc_values = range_list_new();
   if(unalloc_values == NULL)
     return 10;
@@ -770,21 +817,35 @@ int main(int argc, char** argv)
     return ret;
   }
 
-  ret = extractValueLists(f, unalloc_cells, unalloc_keys);
+  ret = extractValueLists(f, unalloc_cells, unalloc_keys,unalloc_linked_values);
   if(ret != 0)
   {
     fprintf(stderr, "ERROR: extractValueLists() failed with %d.\n", ret);
     return ret;
   }
 
-  /* Carve any orphan values and associated data */
-  ret = extractValues(f, unalloc_cells, unalloc_values);
+  /* Carve any orphan values */
+  ret = extractVKs(f, unalloc_cells, unalloc_values);
   if(ret != 0)
   {
-    fprintf(stderr, "ERROR: extractValues() failed with %d.\n", ret);
+    fprintf(stderr, "ERROR: extractVKs() failed with %d.\n", ret);
     return ret;
   }
 
+  /* Carve any data associated with VK records */
+  ret = extractDataCells(f, unalloc_cells, unalloc_linked_values);
+  if(ret != 0)
+  {
+    fprintf(stderr, "ERROR: extractDataCells() failed with %d.\n", ret);
+    return ret;
+  }
+  ret = extractDataCells(f, unalloc_cells, unalloc_values);
+  if(ret != 0)
+  {
+    fprintf(stderr, "ERROR: extractDataCells() failed with %d.\n", ret);
+    return ret;
+  }
+  
   /* Carve any SK records */
   ret = extractSKs(f, unalloc_cells, unalloc_sks);
   if(ret != 0)
@@ -804,7 +865,7 @@ int main(int argc, char** argv)
   for(i=0; i < num_unalloc_keys; i++)
   {
     cur_elem = range_list_get(unalloc_keys, i);
-    tmp_key = (REGF_NK_REC*)cur_elem->data;
+    tmp_key = (REGFI_NK_REC*)cur_elem->data;
 
     if(tmp_key == NULL)
       return 20;
@@ -815,11 +876,10 @@ int main(int argc, char** argv)
   }
   
   /* Now start the output */
-
   for(i=0; i < num_unalloc_keys; i++)
   {
     cur_elem = range_list_get(unalloc_keys, i);
-    tmp_key = (REGF_NK_REC*)cur_elem->data;
+    tmp_key = (REGFI_NK_REC*)cur_elem->data;
 
     printKey(f, tmp_key, parent_paths[i]);
     if(tmp_key->num_values > 0 && tmp_key->values != NULL)
@@ -827,11 +887,18 @@ int main(int argc, char** argv)
       tmp_name = quote_string(tmp_key->keyname, key_special_chars);
       tmp_path = (char*)malloc(strlen(parent_paths[i])+strlen(tmp_name)+2);
       if(tmp_path == NULL)
-	return 10;
-      sprintf(tmp_path, "%s/%s", parent_paths[i], tmp_name);
-      for(j=0; j < tmp_key->num_values; j++)
       {
-	tmp_value = tmp_key->values[j];
+	free(tmp_name);
+	return 10;
+      }
+
+      sprintf(tmp_path, "%s/%s", parent_paths[i], tmp_name);
+      for(j=0; j < tmp_key->values->num_values; j++)
+      {
+	tmp_value = 
+	  (REGFI_VK_REC*)range_list_find_data(unalloc_linked_values, 
+					      tmp_key->values->elements[j]
+					      + REGFI_REGF_SIZE);
 	if(tmp_value != NULL)
 	  printValue(f, tmp_value, tmp_path);
       }
@@ -846,7 +913,7 @@ int main(int argc, char** argv)
   for(i=0; i < range_list_size(unalloc_values); i++)
   {
     cur_elem = range_list_get(unalloc_values, i);
-    tmp_value = (REGF_VK_REC*)cur_elem->data; 
+    tmp_value = (REGFI_VK_REC*)cur_elem->data; 
 
     printValue(f, tmp_value, "");
   }
@@ -860,5 +927,11 @@ int main(int argc, char** argv)
     }
   }
 
+  range_list_free(unalloc_cells);
+  range_list_free(unalloc_keys);
+  range_list_free(unalloc_linked_values);
+  range_list_free(unalloc_values);
+  range_list_free(unalloc_sks);
+
   return 0;
 }
